using Adaptation.Shared;
using System;
using System.Data.SqlClient;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.Json;

namespace Adaptation.FileHandlers.OpenInsight;

public class FromIQS
{

#nullable enable

    private static string GetCommandText(Logistics logistics, pcl.Description description, string dateTime, string? sid)
    {
        StringBuilder result = new();
        _ = result
                .AppendLine(" select ")
                .AppendLine("       se.f_sgrp se_sgrp, ")
                .AppendLine("       se.f_sgtm se_sgtm, ")
                .AppendLine("       se.f_tsno se_tsno, ")
                .AppendLine("       se.f_val se_val, ")
                .AppendLine("       rd.f_name rd_name, ")
                .AppendLine("       jd.f_name jd_name, ")
                .AppendLine("       pl.f_name pl_name, ")
                .AppendLine("       pd.f_name pd_name, ")
                .AppendLine("       td.f_test td_test, ")
                .AppendLine("       td.f_name td_name ")
                .AppendLine(" from [spcepiworld].[dbo].[sgrp_ext] se ")
                .AppendLine(" join [spcepiworld].[dbo].[prcs_dat] rd ")
                .AppendLine("   on se.f_prcs = rd.f_prcs ")
                .AppendLine(" join [spcepiworld].[dbo].[job_dat] jd ")
                .AppendLine("   on se.f_job = jd.f_job ")
                .AppendLine(" join [spcepiworld].[dbo].[part_lot] pl ")
                .AppendLine("   on se.f_lot = pl.f_lot ")
                .AppendLine(" join [spcepiworld].[dbo].[part_dat] pd ")
                .AppendLine("   on se.f_part = pd.f_part ")
                .AppendLine(" join [spcepiworld].[dbo].[test_dat] td  ")
                .AppendLine("   on se.f_test = td.f_test ")
                .AppendLine(" where se.f_flag = 0 ");
        if (!string.IsNullOrEmpty(sid))
            _ = result.Append("   and se.f_sgrp = ").Append(sid).AppendLine("  ");
        _ = result
                .Append("   and rd.f_name = '").Append(description.Reactor).AppendLine("' ")
                .Append("   and pd.f_name = '").Append(description.PSN).AppendLine("' ")
                .AppendLine("   and jd.f_name in ('TENCOR1', 'TENCOR2', 'TENCOR3') ")
                .Append("   and jd.f_name = '").Append(logistics.MesEntity).AppendLine("' ")
                .Append("   and pl.f_name = '").Append(description.RDS).AppendLine("' ")
                .Append("   and dateadd(HH, -7, (dateadd(SS, convert(bigint, se.f_sgtm), '19700101'))) = '").Append(dateTime).AppendLine("' ")
                .AppendLine("   for json path ");
        return result.ToString();
    }

    private static StringBuilder GetForJsonPath(string connectionString, string commandText)
    {
        StringBuilder stringBuilder = new();
        using (SqlConnection sqlConnection = new(connectionString))
        {
            sqlConnection.Open();
            using SqlCommand sqlCommand = new(commandText, sqlConnection);
            SqlDataReader sqlDataReader = sqlCommand.ExecuteReader(System.Data.CommandBehavior.SequentialAccess);
            while (sqlDataReader.Read())
                _ = stringBuilder.Append(sqlDataReader.GetString(0));
        }
        return stringBuilder;
    }

    internal static string GetCommandText(string connectionString, Logistics logistics, pcl.Description description)
    {
        string result;
        string dateTime;
        string dateFormat = pcl.Description.GetDateFormat();
        if (DateTime.TryParseExact(description.Date, dateFormat, CultureInfo.InvariantCulture, DateTimeStyles.None, out DateTime dateTimeParsed))
            dateTime = dateTimeParsed.ToString("yyyy-MM-dd HH:mm:ss");
        else if (DateTime.TryParse(description.Date, CultureInfo.InvariantCulture, DateTimeStyles.None, out dateTimeParsed))
            dateTime = dateTimeParsed.ToString("yyyy-MM-dd HH:mm:ss");
        else
            dateTime = logistics.DateTimeFromSequence.ToString("yyyy-MM-dd HH:mm:ss");
        string commandText = GetCommandText(logistics, description, dateTime, sid: null);
        StringBuilder stringBuilder = GetForJsonPath(connectionString, commandText);
        if (stringBuilder.Length == 0)
            result = commandText;
        else
        {
            JsonElement[]? jsonElements = JsonSerializer.Deserialize<JsonElement[]>(stringBuilder.ToString());
            if (jsonElements is null || !jsonElements.Any() || jsonElements[0].ValueKind != JsonValueKind.Object)
                result = commandText;
            else
            {
                JsonProperty[] jsonProperties = jsonElements[0].EnumerateObject().ToArray();
                if (!jsonProperties.Any() || jsonProperties[0].Name != "se_sgrp")
                    result = commandText;
                else
                    result = GetCommandText(logistics, description, dateTime, sid: jsonProperties[0].Value.ToString());
            }
        }
        return result;
    }

#nullable disable

}